@@ -17,24 +17,32 @@ module Database.Postgres.Temp.Internal.Config where
 import Database.Postgres.Temp.Internal.Core
 
 import           Control.Applicative.Lift
+import           Control.DeepSeq
 import           Control.Exception
 import           Control.Monad (join)
+import           Crypto.Hash.SHA1 (hash)
 import           Control.Monad.Trans.Class
 import           Control.Monad.Trans.Cont
+import qualified Data.ByteString.Char8 as BSC
+import qualified Data.ByteString.Base64.URL as Base64
+import           Data.Char
 import qualified Data.Map.Strict as Map
 import           Data.Map.Strict (Map)
 import           Data.Maybe
 import           Data.Monoid
 import           Data.Monoid.Generic
+import           Data.List
 import qualified Database.PostgreSQL.Simple.Options as Client
 import           GHC.Generics (Generic)
 import           Network.Socket.Free (getFreePort)
 import           System.Directory
 import           System.Environment
+import           System.Exit (ExitCode(..))
 import           System.IO
 import           System.IO.Error
 import           System.IO.Temp (createTempDirectory)
 import           System.IO.Unsafe (unsafePerformIO)
+import           System.Process
 import           Text.PrettyPrint.ANSI.Leijen hiding ((<$>))
 
 prettyMap :: (Pretty a, Pretty b) => Map a b -> Doc
@@ -202,14 +210,14 @@ standardProcessConfig = mempty
   , stdErr = pure stderr
   }
 
--- | A global reference to @/dev/null@ 'Handle'.
+-- | A global reference to @\/dev\/null@ 'Handle'.
 --
 --   @since 1.12.0.0
 devNull :: Handle
 devNull = unsafePerformIO (openFile "/dev/null" WriteMode)
 {-# NOINLINE devNull #-}
 
--- | 'silentProcessConfig' sets the handles to @/dev/null@ and
+-- | 'silentProcessConfig' sets the handles to @\/dev\/null@ and
 --   inherits the environment variables from the calling process.
 --
 --   @since 1.12.0.0
@@ -256,7 +264,8 @@ completeProcessConfig envs ProcessConfig {..} = runErrors $ do
 --
 --   @since 1.12.0.0
 data CompleteDirectoryType = CPermanent FilePath | CTemporary FilePath
-  deriving(Show, Eq, Ord)
+  deriving stock (Show, Eq, Ord, Generic)
+  deriving anyclass (NFData)
 
 -- | Get the file path of a 'CompleteDirectoryType', regardless if it is a
 -- 'CPermanent' or 'CTemporary' type.
@@ -291,8 +300,8 @@ data DirectoryType
 
 instance Pretty DirectoryType where
   pretty = \case
-    Permanent x -> text "CPermanent" <+> pretty x
-    Temporary   -> text "CTemporary"
+    Permanent x -> text "Permanent" <+> pretty x
+    Temporary   -> text "Temporary"
 
 -- | Takes the last 'Permanent' value.
 instance Semigroup DirectoryType where
@@ -317,7 +326,11 @@ setupDirectoryType
   -> IO CompleteDirectoryType
 setupDirectoryType tempDir pat = \case
   Temporary -> CTemporary <$> createTempDirectory tempDir pat
-  Permanent x  -> pure $ CPermanent x
+  Permanent x  -> CPermanent <$> case x of
+    '~':rest -> do
+      homeDir <- getHomeDirectory
+      pure $ homeDir <> "/" <> rest
+    xs -> pure xs
 
 -- Remove a temporary directory and ignore errors
 -- about it not being there.
@@ -383,6 +396,7 @@ completePostgresPlan envs PostgresPlan {..} = runErrors $ do
 data Plan = Plan
   { logger :: Last Logger
   , initDbConfig :: Maybe ProcessConfig
+  , copyConfig :: Last (Maybe CopyDirectoryCommand)
   , createDbConfig :: Maybe ProcessConfig
   , postgresPlan :: PostgresPlan
   , postgresConfigFile :: [String]
@@ -404,6 +418,9 @@ instance Pretty Plan where
     <> text "initDbConfig:"
     <> softline
     <> indent 2 (pretty createDbConfig)
+    <> text "copyConfig:"
+    <> softline
+    <> indent 2 (pretty (getLast copyConfig))
     <> hardline
     <> text "postgresPlan:"
     <> softline
@@ -423,6 +440,7 @@ completePlan envs Plan {..} = runErrors $ do
   completePlanLogger   <- getOption "logger" logger
   completePlanInitDb   <- eitherToErrors $ addErrorContext "initDbConfig: " $
     traverse (completeProcessConfig envs) initDbConfig
+  completePlanCopy   <- getOption "dataDirectoryString" copyConfig
   completePlanCreateDb <- eitherToErrors $ addErrorContext "createDbConfig: " $
     traverse (completeProcessConfig envs) createDbConfig
   completePlanPostgres <- eitherToErrors $ addErrorContext "postgresPlan: " $
@@ -463,6 +481,7 @@ data Config = Config
   , temporaryDirectory :: Last FilePath
   -- ^ The directory used to create other temporary directories. Defaults
   --   to @/tmp@.
+  , initDbCache :: Last (Maybe (Bool, FilePath))
   }
   deriving stock (Generic)
   deriving Semigroup via GenericSemigroup Config
@@ -484,9 +503,11 @@ instance Pretty Config where
     <> hardline
     <> text "port:" <+> pretty (getLast port)
     <> hardline
-    <> text "dataDirectory:"
+    <> text "temporaryDirectory:"
     <> softline
     <> pretty (getLast temporaryDirectory)
+    <> hardline
+    <> text "initDbCache:" <+> pretty (getLast initDbCache)
 
 socketDirectoryToConfig :: FilePath -> [String]
 socketDirectoryToConfig dir =
@@ -494,21 +515,116 @@ socketDirectoryToConfig dir =
     , "unix_socket_directories = '" <> dir <> "'"
     ]
 
+-------------------------------------------------------------------------------
+-- Caching
+-------------------------------------------------------------------------------
+getInitDbVersion :: IO String
+getInitDbVersion = readProcessWithExitCode "initdb" ["--version"] "" >>= \case
+  (ExitSuccess, outputString, _) -> do
+    let
+      theLastPart = last $ words outputString
+      versionPart = takeWhile (\x -> isDigit x || x == '.' || x == '-') theLastPart
+    pure $ if last versionPart == '.'
+             then init versionPart
+             else versionPart
+
+  (startErrorExitCode, startErrorStdOut, startErrorStdErr) ->
+    throwIO InitDbFailed {..}
+
+makeCommandLine :: String -> CompleteProcessConfig -> String
+makeCommandLine command CompleteProcessConfig {..} =
+  let envs = unwords $ map (\(x, y) -> x <> "=" <> y) completeProcessConfigEnvVars
+      args = unwords completeProcessConfigCmdLine
+  in envs <> " " <> command <> args
+
+makeInitDbCommandLine :: CompleteProcessConfig -> String
+makeInitDbCommandLine = makeCommandLine "initdb"
+
+makeArgumentHash :: String -> String
+makeArgumentHash = BSC.unpack . Base64.encode . hash . BSC.pack
+
+makeCachePath :: FilePath -> String -> IO String
+makeCachePath cacheFolder cmdLine = do
+  version <- getInitDbVersion
+  let theHash = makeArgumentHash cmdLine
+  pure $ cacheFolder <> "/" <> version <> "/" <> theHash
+
+
+splitDataDirectory :: CompleteProcessConfig -> (Maybe String, CompleteProcessConfig)
+splitDataDirectory old =
+  let isDataDirectoryFlag xs = "-D" `isPrefixOf` xs || "--pgdata=" `isPrefixOf` xs
+      (dataDirectoryArgs, otherArgs) =
+        partition isDataDirectoryFlag $ completeProcessConfigCmdLine old
+
+      firstDataDirectoryArg = flip fmap (listToMaybe dataDirectoryArgs) $ \case
+        '-':'D':' ':theDir -> theDir
+        '-':'D':theDir -> theDir
+        '-':'-':'p':'g':'d':'a':'t':'a':'=':theDir -> theDir
+        _ -> error "splitDataDirectory not possible"
+
+      filteredEnvs = filter (not . ("PGDATA"==) . fst) $
+        completeProcessConfigEnvVars old
+
+      clearedConfig = old
+        { completeProcessConfigCmdLine = otherArgs
+        , completeProcessConfigEnvVars = filteredEnvs
+        }
+
+  in (firstDataDirectoryArg, clearedConfig)
+
+addDataDirectory :: String -> CompleteProcessConfig -> CompleteProcessConfig
+addDataDirectory theDataDirectory x = x
+  { completeProcessConfigCmdLine =
+      ("--pgdata=" <> theDataDirectory) : completeProcessConfigCmdLine x
+  }
+
+cachePlan :: CompletePlan -> Bool -> FilePath -> IO CompletePlan
+cachePlan plan@CompletePlan {..} cow cacheDirectory = case completePlanInitDb of
+  Nothing -> pure plan
+  Just theConfig -> do
+    let (mtheDataDirectory, clearedConfig) = splitDataDirectory theConfig
+    theDataDirectory <- maybe
+      (throwIO $ FailedToFindDataDirectory (show $ pretty clearedConfig))
+      pure
+      mtheDataDirectory
+
+    let theCommandLine = makeInitDbCommandLine clearedConfig
+
+    cachePath <- makeCachePath cacheDirectory theCommandLine
+
+    let cachedDataDirectory = cachePath <> "/data"
+
+    theInitDbPlan <- doesDirectoryExist cachePath >>= \case
+      True -> pure $ Nothing
+      False -> do
+        createDirectoryIfMissing True cachePath
+        writeFile (cachePath <> "/commandLine.log") theCommandLine
+        pure $ pure $ addDataDirectory cachedDataDirectory clearedConfig
+
+    pure plan
+      { completePlanCopy = pure $ CopyDirectoryCommand
+        { copyDirectoryCommandSrc = cachedDataDirectory
+        , copyDirectoryCommandDst = theDataDirectory
+        , copyDirectoryCommandCow = cow
+        }
+      , completePlanInitDb = theInitDbPlan
+      }
+
 -- | Create a 'Plan' that sets the command line options of all processes
 --   (@initdb@, @postgres@ and @createdb@). This the @generated@ plan
 --   that is combined with the @extra@ plan from
 --   'Database.Postgres.Temp.startConfig'.
 toPlan
   :: Bool
-  -- ^ Make @initdb@ options
+  -- ^ Make @initdb@ options.
   -> Bool
-  -- ^ Make @createdb@ options
+  -- ^ Make @createdb@ options.
   -> Int
-  -- ^ port
+  -- ^ The port.
   -> FilePath
-  -- ^ Socket directory
+  -- ^ Socket directory.
   -> FilePath
-  -- ^ The @postgres@ data directory
+  -- ^ The @postgres@ data directory.
   -> Plan
 toPlan makeInitDb makeCreateDb port socketDirectory dataDirectoryString = mempty
   { postgresConfigFile = socketDirectoryToConfig socketDirectory
@@ -546,12 +662,11 @@ toPlan makeInitDb makeCreateDb port socketDirectory dataDirectoryString = mempty
             { keyBased = Map.fromList
                 [("--pgdata=", Just dataDirectoryString)]
             }
-
         }
       else Nothing
+  , copyConfig = pure Nothing
   }
 
-
 -- | Create all the temporary resources from a 'Config'. This also combines the
 -- 'Plan' from 'toPlan' with the @extra@ 'Config' passed in.
 setupConfig
@@ -562,6 +677,7 @@ setupConfig Config {..} = evalContT $ do
   envs <- lift getEnvironment
   thePort <- lift $ maybe getFreePort pure $ join $ getLast port
   let resourcesTemporaryDir = fromMaybe "/tmp" $ getLast temporaryDirectory
+      resourcesInitDbCache = join $ getLast initDbCache
   resourcesSocketDirectory <- ContT $ bracketOnError
     (setupDirectoryType resourcesTemporaryDir "tmp-postgres-socket" socketDirectory) cleanupDirectoryType
   resourcesDataDir <- ContT $ bracketOnError
@@ -573,9 +689,10 @@ setupConfig Config {..} = evalContT $ do
         (toFilePath resourcesSocketDirectory)
         (toFilePath resourcesDataDir)
       finalPlan = hostAndDir <> plan
-  resourcesPlan <- lift $
+  uncachedPlan <- lift $
     either (throwIO . CompletePlanFailed (show $ pretty finalPlan)) pure $
       completePlan envs finalPlan
+  resourcesPlan <- lift $ maybe (pure uncachedPlan) (uncurry $ cachePlan uncachedPlan) resourcesInitDbCache
   pure Resources {..}
 
 -- | Free the temporary resources created by 'setupConfig'.
@@ -605,6 +722,7 @@ data Resources = Resources
   , resourcesTemporaryDir :: FilePath
   -- ^ The directory where other temporary directories are created.
   --   Usually @/tmp.
+  , resourcesInitDbCache :: Maybe (Bool, FilePath)
   }
 
 instance Pretty Resources where
